// Copyright (c) 2017, Ravi Teja Gudapati. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:math';
import 'dart:async';
import 'package:jaguar/jaguar.dart';
import 'package:jaguar_cache/jaguar_cache.dart';

final Cache defaultThrottleCache = new InMemoryCache(new Duration(seconds: -1));

/// Contains count information for a single throttle identifier/key
class ThrottleData {
  /// Count
  int count = 0;

  /// Start of the recorded time interval
  int milliseconds = 0;

  ThrottleData();

  ThrottleData.now(this.count)
      : milliseconds = new DateTime.now().millisecondsSinceEpoch;

  ThrottleData.make(this.count, this.milliseconds);

  /// Returns time
  DateTime time() => new DateTime.fromMillisecondsSinceEpoch(milliseconds);

  /// Increments the count
  ThrottleData inc(int milliseconds) =>
      new ThrottleData.make(count + 1, milliseconds);
}

/// Identifies a HTTP request given the [Context] object
///
/// For example identifier could be based on
///   1. IP address
///   2. Logged-in User
///   3. Session id
///   4. route
///   5. Etc
///   6. Or any combination of the above
typedef FutureOr<String> ThrottleIdMaker(Context ctx);

/// [ThrottleIdMaker] to throttle requests by remote IP
String throttleIdByIp(Context ctx) =>
    ctx.req.connectionInfo.remoteAddress.address;

/// Interceptor to rate-limit routes
class Throttler extends Interceptor {
  /// Maximum number of requests allowed for the given [interval]
  final Rate quota;

  /// Identification maker. Creates an identification for the given request
  final ThrottleIdMaker idMaker;

  final Cache store;

  /// Provides an opportunity to create [Response] when rate-limit exceeds
  final RouteHandlerFunc failResponse;

  Throttler(this.quota,
      {ThrottleIdMaker idMaker, this.failResponse, Cache store})
      : store = store ?? defaultThrottleCache,
        idMaker = idMaker ?? throttleIdByIp;

  factory Throttler.perSec(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perSec(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  factory Throttler.perMin(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perMin(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  factory Throttler.perHour(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perHour(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  factory Throttler.perDay(int quota,
          {ThrottleIdMaker idMaker,
          RouteHandlerFunc failResponse,
          Cache store}) =>
      new Throttler(perDay(quota),
          idMaker: idMaker, failResponse: failResponse, store: store);

  ThrottleState _state;

  Future<ThrottleState> pre(Context ctx) async {
    final String throttleId = await idMaker(ctx);
    final ThrottleData count = await _visit(throttleId, quota.interval);

    _state = new ThrottleState.make(quota.count, count, quota.interval);

    if (count == null) return _state;

    if (count.count > quota.count) {
      final Response resp = failResponse != null
          ? await failResponse(ctx)
          : new Response('Limit exceeded', statusCode: 429);

      throw _state.setHeaders(resp);
    }

    return _state;
  }

  Response post(Context ctx, Response incoming) => _state.setHeaders(incoming);

  Future<ThrottleData> _visit(String identifier, Duration interval) async {
    ThrottleData ret = new ThrottleData.now(1);

    ThrottleData data;

    try {
      data = await store.read(identifier);
    } catch (e) {
      if (e != cacheMiss) rethrow;
    }

    final now = new DateTime.now();

    if (data == null) {
      ret = new ThrottleData.now(1);
    } else if (now.difference(data.time()) < interval) {
      ret = data.inc(data.milliseconds);
    } else {
      ret = new ThrottleData.now(1);
    }

    await store.upsert(identifier, ret);
    return ret;
  }
}

/// Contains the throttle state information
class ThrottleState {
  /// Max quota
  final int quota;

  /// Remaining quota
  final int remaining;

  /// Duration after which quota resets
  final int resetsIn;

  /// Duration after which client can retry in case quota is fully used
  final int retryAfter;

  ThrottleState(this.quota, this.remaining, this.resetsIn, this.retryAfter);

  factory ThrottleState.make(int quota, ThrottleData data, Duration interval) {
    if (data == null) {
      return new ThrottleState(quota, quota, interval.inSeconds, -1);
    }

    int remaining = max(0, quota - data.count);
    int resetsIn = max(0, data.time().difference(new DateTime.now()).inSeconds);
    int retryAfter = -1;
    if (quota <= data.count) {
      retryAfter = max(0,
          data.time().add(interval).difference(new DateTime.now()).inSeconds);
    }
    return new ThrottleState(quota, remaining, resetsIn, retryAfter);
  }

  Response setHeaders(Response resp) {
    resp.headers.set('X-RateLimit-Limit', quota.toString());
    resp.headers.set('X-RateLimit-Remaining', remaining.toString());
    resp.headers.set('X-RateLimit-Reset', resetsIn.toString());
    if (retryAfter >= 0) resp.headers.set('Retry-After', retryAfter.toString());
    return resp;
  }
}

/// Rate structure. Holds [count] per [interval]
class Rate {
  /// The time duration of the interval
  final Duration interval;

  /// Count per interval
  final int count;

  const Rate(this.interval, this.count);

  /// Creates [Rate] with [count] per second
  const Rate.perSec(this.count) : interval = const Duration(seconds: 1);

  /// Creates [Rate] with [count] per minute
  const Rate.perMin(this.count) : interval = const Duration(minutes: 1);

  /// Creates [Rate] with [count] per hour
  const Rate.perHour(this.count) : interval = const Duration(hours: 1);

  /// Creates [Rate] with [count] per day
  const Rate.perDay(this.count) : interval = const Duration(days: 1);
}

/// Creates [Rate] with [count] per second
Rate perSec(int count) => new Rate.perSec(count);

/// Creates [Rate] with [count] per minute
Rate perMin(int count) => new Rate.perMin(count);

/// Creates [Rate] with [count] per hour
Rate perHour(int count) => new Rate.perHour(count);

/// Creates [Rate] with [count] per day
Rate perDay(int count) => new Rate.perDay(count);

/// Throttlers for given [id] with [quota]
Future<ThrottleState> throttle(String id, Rate quota, {Cache store}) async {
  store ??= defaultThrottleCache;
  final ThrottleData count = await _visit(id, quota.interval, store);

  final state = new ThrottleState.make(quota.count, count, quota.interval);

  if (count == null) return state;

  if (count.count > quota.count) throw state;

  return state;
}

Future<ThrottleData> _visit(
    String identifier, Duration interval, Cache store) async {
  ThrottleData ret = new ThrottleData.now(1);

  ThrottleData data;

  try {
    data = await store.read(identifier);
  } catch (e) {
    if (e != cacheMiss) rethrow;
  }

  final now = new DateTime.now();

  if (data == null) {
    ret = new ThrottleData.now(1);
  } else if (now.difference(data.time()) < interval) {
    ret = data.inc(data.milliseconds);
  } else {
    ret = new ThrottleData.now(1);
  }

  await store.upsert(identifier, ret);
  return ret;
}
